:- use_module(library(lists)).
:- use_module(kkigyo).
:- use_module(backtrack_count).

% :- type kigyotabla      ---> k(meret, pont, pont, list(ref_pont)).
% :- type ref_pont        ---> p(sor, oszlop, ertek).
% :- type sor               == integer.
% :- type oszlop            == integer.
% :- type ertek             == integer.
% :- type pont            ---> sor-oszlop.
% :- type meret           ---> sor-oszlop.
% :- type kigyovonal        == list(pont).


% A Feladvany kigyo-feladvany megoldasa a KV kigyovonal.
% :- pred kigyo(kigyotabla::in, kigyovonal::out).
kigyo(Feladvany, KV) :-
	korlatok_keszitese(Feladvany, PMx, KorlL),
	kovetkeztetes(KorlL, []),
	kigyovonal_eloallitasa(Feladvany, PMx, KV).


% Korlatok elkeszitese
% --------------------

% A kigyo feladvany megoldasa harom fazisbol all:
%    1. az un. korlatok (constraint-ek) eloallitasa
%    2. a korlatokon valo kovetkeztetes
%    3. a kimenet kivant alakjanak (kigyovonal) eloallitasa
% A fenti fazisok kozul csak a 2. jar keresessel, az 1. es a
% 3. determinisztikus.

% A feladat a kigyodarabok helyenek meghatarozasa egy adott meretu
% teglalapban (matrixban). A matrix minden elemehez egy 0-1 erteku valtozot
% rendelunk, ahol az 1 ertek jelenti a kigyodarabot (ez az un.
% pontmatrix). A korlatok ezen valtozok kozotti osszefuggeseket irnak le:

%  sum(L, N) (osszegkorlat):
%    Az L listaban levo valtozok osszege N.
%  csum(P, L, N) (felteteles osszegkorlat):
%    Ha P = 1, akkor az L listaban levo valtozok osszege N.

% Pelda

% A feladvany: k(2-3,1-1,1-3,[p(1,2,5)]).
% A pontmatrix:

%        P11  P12  P13
%        P21  P22  P23

% Az osszegkorlatok:

%        sum([P12,P21],1)     % a kezdo- ill. vegpont elszomszedjai kozott
%        sum([P12,P23],1)     % pontosan egy kigyodarab van.
%        sum([P11,P21,P22,P23,P13],5)
%                             % a referenciapont el- es sarokszomszedai
%                             % kozott a megadott szamu kigyodarab van.
%        csum(P21,[P11,P22],2)     % ha valamely masik ponton kigyodarab van
%        csum(P22,[P11,P12,P23],2) % akkor pontosan 2 elszomszedjan
%        csum(P23,[P22,P13],2)     % van kigyodarab

% A kormentesseg biztositasara (hogy a megoldasban ne lehessenek korok) egy
% masik valtozotablat (az un. kapcsolatmatrixot) es egy ujabb korlatfajtat
% hasznalunk. A kapcsolatmatrixban levo ket (kapcsolati) valtozo csakkor
% lesz azonos, ha a nekik megfelelo pontok kozott kialakult mar egy
% folytonos kigyoresz.

% Minden elszomszedos pontparra felirunk egy kapcsolati korlatot, amelynek
% argumentumai a ket pontnak megfelelo pontvaltozo (Pij, Pkl) es kapcsolati
% valtozo (Kij,Kkl).

% A peldaban:

% A kapcsolatmatrix:

%        K11  K12  K13
%        K21  K22  K23

% A kapcsolati korlatok:

%        edge(P11,P12,K11,K12)
%        edge(P11,P21,K11,K21)
%        edge(P12,P13,K12,K13)
%        edge(P12,P22,K12,K22)
%        edge(P13,P23,K13,K23)
%
%        edge(P21,P21,K21,K21)
%        edge(P22,P23,K22,K23)

% A kapcsolati korlatnak, az osszegkorlatokkal ellentetben, csak
% proceduralis szemantikaja van:

%  edge(P1,P2,K1,K2):
%    - Ha P1 vagy P2 0, akkor az edge korlat feltetel nelkul fennal.
%    - Ha P1 es P2 1 (azaz mindket ponton kigyodarab van), akkor a ket
%      kapcsolati valtozot egyesitjuk egymassal: K1 = K2.
%    - Mielott a fenti egyesitest elvegeznenk, megvizsgaljuk, hogy fennall-e
%      a K1==K2 azonossag (azaz a ket pont *kozvetve* mar ossze volt kotve a
%      kigyo altal), es ez esetben meghiusulunk

% :- type korlat ---> sum(list(kpont), int) |
%                     csum(kpont, list(kpont), int) |
%                     edge(kpont, kpont, var, var).
% :- type kpont ==    int.        % 0 vagy 1 ertek lehet csak

% :- type matrix(T) == list(list(T)).
% Megjegyzes: a matrixban a sorokat reprezentalo listak azonos
% hosszusaguak.

% A Feladvany-hoz tartozo pontmatrix PMx, a neki megfelelo
% korlatok listaja KorlL.
% :- pred korlatok_keszitese(kigyotabla::in,
%                            matrix(kpont)::out,
%                            list(korlat)::out).
korlatok_keszitese(Feladvany, PMx, KorlL) :-
	Feladvany = k(MaxS-MaxO,_,_,_),
	matrix_of_vars(MaxS, MaxO, PMx),
	matrix_of_vars(MaxS, MaxO, KMx),
	korlatok(Feladvany, PMx, KMx, KorlL, []).

% korlatok(Feladvany, PMx, KMx, KL0, KL):
% A Feladvany-hoz tartozo pontmatrix PMx, a hozza tartozo kapcsolati
% matrix KMx, es a neki megfelelo korlatok listaja KL0-KL.
% :- pred korlatok(kigyotabla::in, list(kpont)::in, list(var)::in,
%                  list(korlat)::out, list(korlat)::in).
korlatok(Feladvany, PMx, KMx) -->
	{Feladvany = k(_,KP,VP,RL)},
	{egyeb_pontok(Feladvany, EL)},
	kezdo_veg_korlat(KP, PMx),
	kezdo_veg_korlat(VP, PMx),
	refpont_korlatok(RL, PMx),
	egyeb_pont_korlatok(EL, PMx),
	kapcsolati_korlatok([KP,VP|EL], PMx, KMx).

% egyeb_pontok(Feladvany, EL): A Feladvany tablajaban levo, a kezdo-, veg-
% es referencia-pontoktol kulonbozo pontok (Sor-Oszlop) alaku
% koordinatainak listaja EL.
% :- pred egyeb_pontok(kigyotabla::in, list(pont)::out).
egyeb_pontok(k(MaxS-MaxO,KP,VP,RL), EL) :-
	findall(EP,
		(   between(1, MaxS, S),
		    between(1, MaxO, O),
		    \+ member(p(S,O,_), RL),
		    EP = S-O,
		    EP \== KP,
		    EP \== VP
		),
		EL).

% kezdo_veg_korlat(KP, PMx, KL0, KL):
% A KP kezdo- vagy vegponthoz a PMx pontmatrixban tartozo
% korlatok listaja KL0-KL. A pontmatrixban az adott ponton 1 van.
% :- pred kezdo_veg_korlat(pont::in, matrix(kpont),
%                          list(korlat)::out, list(korlat)::in).
kezdo_veg_korlat(KP, PMx) -->
	{KP = S-O,
	 matrix_elem(S, O, PMx, 1)},
	{elszomszedjai(KP, PMx, L)},
	[sum(L, 1)].

% refpont_korlatok(RL, PMx, KL0, KL):
% Az RL referenciapont-listahoz a PMx pontmatrixban tartozo
% korlatok listaja KL0-KL. A pontmatrixban a referenciapontokban 0 van.
% :- pred refpont_korlatok(list(ref_pont)::in, matrix(kpont),
%                          list(korlat)::out, list(korlat)::in).
refpont_korlatok([], _PMx) --> [].
refpont_korlatok([RP|RL], PMx) -->
	refpont_korlat(RP, PMx),
	refpont_korlatok(RL, PMx).

% refpont_korlat(RP, PMx, KL0, KL):
% Az RP referenciaponthoz a PMx pontmatrixban tartozo korlatok listaja
% KL0-KL. A pontmatrixban a referenciapontban 0 van.
% :- pred refpont_korlatok(ref_pont::in, matrix(kpont),
%                          list(korlat)::out, list(korlat)::in).
refpont_korlat(p(S,O,Sz), PMx) -->
	{matrix_elem(S, O, PMx, 0)},
	{szomszedjai(S-O, PMx, L)},
	[sum(L, Sz)].

% egyeb_pont_korlatok(EL, PMx, KL0, KL):
% Az EL egyebpont-listahoz a PMx pontmatrixban tartozo
% korlatok listaja KL0-KL.
% :- pred egyeb_pont_korlatok(list(pont)::in, matrix(kpont),
%                             list(korlat)::out, list(korlat)::in).
egyeb_pont_korlatok([], _PMx) --> [].
egyeb_pont_korlatok([EP|EL], PMx) -->
	egyeb_pont_korlat(EP, PMx),
	egyeb_pont_korlatok(EL, PMx).

% egyeb_pont_korlat(EP, PMx, KL0, KL):
% Az EP "egyeb" ponthoz a PMx pontmatrixban tartozo
% korlatok listaja KL0-KL.
% :- pred egyeb_pont_korlat(pont::in, matrix(kpont),
%                           list(korlat)::out, list(korlat)::in).
egyeb_pont_korlat(P, PMx) -->
	{P = S-O,
	 matrix_elem(S, O, PMx, B)},
	{elszomszedjai(P, PMx, L)},
	[csum(B, L, 2)].

% kapcsolati_korlatok(PL, PMx, KMx, KL0, KL):
% A PL pont-listahoz a PMx pontmatrix ill. KMx kapcsolati matrix szerint
% tartozo kapcsolati korlatok listaja KL0-KL.
% :- pred kapcsolati_korlatok(list(pont)::in, matrix(kpont), matrix(var),
%                             list(korlat)::out, list(korlat)::in).
kapcsolati_korlatok([], _PMx, _KMx) --> [].
kapcsolati_korlatok([P|PL], PMx, KMx) -->
	pont_kapcsolati_korlatai(P, PMx, KMx),
	kapcsolati_korlatok(PL, PMx, KMx).

% pont_kapcsolati_korlatai(P, PMx, KMx, KL0, KL):
% A P ponthoz a PMx pontmatrix ill. KMx kapcsolati matrix szerint
% tartozo kapcsolati korlatok listaja KL0-KL.
% :- pred pont_kapcsolati_korlatai(pont::in, matrix(kpont), matrix(var),
%                                  list(korlat)::out, list(korlat)::in).
pont_kapcsolati_korlatai(P, PMx, KMx) -->
	{deli_elszomszedja(P, DSz)},
	{nyugati_elszomszedja(P, NySz)},
	{P = S-O,
	 matrix_elem(S, O, PMx, B),
	 matrix_elem(S, O, KMx, BK)},
	kapcsolati_korlat(B, BK, DSz, PMx, KMx),
	kapcsolati_korlat(B, BK, NySz, PMx, KMx).

% kapcsolati_korlat(BP, BK, Sz, PMx, KMx, KL0, KL):
% Egy ponthoz rendelt valtozo BP, a hozza rendelt kapcsolati valtozo BK,
% mig Sz az adott pont egy elszomszedjanak a koordinataparja. Az adott
% ponthoz es adott szomszedjahoz tartozo kapcsolati korlatok listaja KL0-KL,
% a PMx pontmatrix ill. KMx kapcsolati matrix szerint.
% :- pred kapcsolati_korlata(kpont, var, pont::in, matrix(kpont), matrix(var),
%                            list(korlat)::out, list(korlat)::in).
kapcsolati_korlat(B, BK, Sz, PMx, KMx) -->
	{Sz = S-O,
	 matrix_elem(S, O, PMx, SzP),
	 matrix_elem(S, O, KMx, SzK),
	 SzP \== 0}, !,
	[edge(B, SzP, BK, SzK)].
kapcsolati_korlat(_, _, _, _, _) --> [].

% Szomszedsagi viszonyok --- segedeljarasok
% -----------------------------------------

% P pont elszomszedja Sz.
% :- pred elszomszedja(pont::in, pont).
elszomszedja(P, Sz) :-
	abs1(Irany),
	elszomszedja(P, Irany, Sz).

% elszomszedja(P, Irany, Sz): P pont Irany iranyu elszomszedja Sz.
% :- pred elszomszedja(pont::in, int::in, pont).
elszomszedja(S-O, Irany, S-O1) :-
	O1 is O+Irany.
elszomszedja(S-O, Irany, S1-O) :-
	S1 is S+Irany.

% deli_elszomszedja(P, Sz) : P pont deli elszomszedja Sz.
% :- pred deli_elszomszedja(pont::in, pont).
deli_elszomszedja(S-O, S-O1) :-
	O1 is O+1.

% nyugati_elszomszedja(P, Sz) : P pont nyugati elszomszedja Sz.
% :- pred nyugati_elszomszedja(pont::in, pont).
nyugati_elszomszedja(S-O, S1-O) :-
	S1 is S+1.

% abs1(X): X abszolut erteke 1.
% :- pred abs1(int).
abs1(1).
abs1(-1).

% P pont el- vagy sarokszomszedja Sz.
% :- pred szomszedja(pont::in, pont).
szomszedja(P, Sz) :-
	P = S-O,
	Sz = S1-O1,
	absleq1(X),
	absleq1(Y),
	S1 is S+X,
	O1 is O+Y,
	P \== Sz.

% absleq1(X): X egesz abszolut erteke kisebb egyenlo 1-nel.
% :- pred absleq1(int).
absleq1(-1).
absleq1(0).
absleq1(1).

% P elszomszedjai pont-valtozoinak listaja L, a PMx pontmatrix szerint.
% :- pred elszomszedjai(pont::in, matrix(kpont)::in, list(kpont)::out).
elszomszedjai(P, PMx, L) :-
	findall(Sz, elszomszedja(P, Sz), SzL),
	matrix_elems(SzL, PMx, L).

% P szomszedjai pont-valtozoinak listaja L, a PMx pontmatrix szerint.
% :- pred szomszedjai(pont::in, matrix(kpont)::in, list(kpont)::out).
szomszedjai(P, PMx, L) :-
	findall(Sz, szomszedja(P, Sz), SzL),
	matrix_elems(SzL, PMx, L).

% matrix_elems(PL, Mx, EL): Az Mx matrix PL listaban megadott koordinataju
% elemeinek listaja EL.
% :- pred matrix_elems(list(pont), matrix(T), list(T)).
matrix_elems([], _Mx, []).
matrix_elems([S-O|PL], Mx, EL) :-
	(   matrix_elem(S, O, Mx, E) -> EL = [E|EL1]
	;   EL1 = EL
	),
	matrix_elems(PL, Mx, EL1).


% Kigyovonal eloallitasa a kitoltott pontmatrix alapjan
% -----------------------------------------------------

% Ez a kigyo-feladvany megoldasanak 3. fazisa. Mar megvan a megoldas, 0-1
% ertekekkel kitoltott pontmatrix formajaban. Ezt kell a kert kigyovonal
% formatumara alakitani.

% kigyovonal_eloallitasa(Feladvany, PMx, KV) :
% A Feladvany megoldasat tartalmazo PMx pontmatrixnak megfelelo
% kigyovonalat irja le a KV lista.
% :- pred kigyovonal_eloallitasa(kigyotabla::in, matrix(int)::in,
%				 kigyovonal::out).
kigyovonal_eloallitasa(k(_,KP,VP,_), PMx, KV) :-
	kigyovonal(KP, VP, 0-0, PMx, KV).


% kigyovonal(KP, VP, EP, PMx, KV):
% A PMx pontmatrix altal leirt kigyovonal KP es VP pontok kozott halado
% szakasza KV, ahol a kigyovonalon a KP elotti pont EP.
% :- pred kigyovonal(pont::in, pont::in, pont::in, matrix(int)::in,
%                    kigyovonal::out).
kigyovonal(KP, KP, _EP, _PMx, KL) :- !,
	KL = [KP].
kigyovonal(KP, VP, EP, PMx, [KP|KL]) :-
	elszomszedja(KP, P),
	EP \== P,
	kigyopont(P, PMx), !,
	kigyovonal(P, VP, KP, PMx, KL).

% kigyopont(P, PMx): A P pont a PMx pontmatrix szerint egy kigyopont.
% :- pred kigyopont(pont::in, matrix(int)::in).
kigyopont(S-O, PMx) :-
	matrix_elem(S, O, PMx, 1).


% Kovetkeztetes
% -------------

% Ez a kigyo-feladvany megoldasanak 2. fazisa, amely a megoldas kereseset
% valositja meg. Belepesi pontja a kovetkeztetes/2 eljaras, amelyet
% kovetkeztetes(KL0, []) alakban hivunk meg. Itt KL0 a feladvanybol
% generalt korlatok listaja, amelyek argumentumaiban a pontmatrix es a
% kapcsolati matrix valtozoi talalhatok.

% A kovetkeztetes alaplepese az, hogy a korlatlistat probaljuk
% egyszerusiteni. Ennek lehetseges modjai:

%    - egy korlatot egy vele ekvivalens egyszerubb korlatta alakitunk, pl.:
%             sum([1,X,1,0,Y], 3) ===> sum([X,Y], 1)
%             csum(1, [X,Y,Z], 1) ===> sum([X,Y,Z], 1)

%    - ha egy korlat azonosan hamis, akkor meghiusulunk, pl.:
%             sum([1,1,X], 1) ===>  false
%             edge(1, 1, K, K) ===> false

%    - ha egy korlat azonosan igaz, akkor elhagyjuk a korlatlistabol, pl.:
%             sum([1,1], 2) ===>  true
%             csum(0, [X,Y,Z], 2) ===> true
%             edge(0, _, _, _) ===> true

%    - ha egy korlat csak egyfele modon teljesulhet, akkor az ehhez
%      szukseges valtozo-behelyettesiteseket elvegezzuk es a korlatot
%      elhagyjuk, pl.:

%             sum([X,Y], 2) ===> X = 1, Y = 1
%             sum([X,Y,Z], 0) ===> X = 0, Y = 0, Z = 0
%             csum(P, [0,0,X], 2) ===> P = 0
%             edge(1, 1, K1, K2) ===> K1 \== K2, K1 = K2

% A fenti lepeseket mindaddig ismeteljuk, amig ujabb egyszerusitesek
% lehetsegesek. Ha a korlatlista mar tovabb nem egyszerusitheto, akkor
% valasztasi pontot csinalunk, oly modon, hogy egy (jol megvalasztott)
% pontvaltozot 0-ra ill. 1-re helyettesitunk, majd folytatjuk az
% egyszerusiteseket. Ha ily modon el tudunk jutni egy ures korlatlistahoz,
% akkor megoldottuk a feladatot, es kozben a pontvaltozok mind 0-1
% ertekekre helyettesitodtek.


% kovetkeztetes(KL0, KL): A KL0 korlatlista KL-le egyszerusitheto, esetleg
% valasztasi pontok letrehozasaval.
% :- pred kovetkeztetes(list(korlat)::in, list(korlat)::in).
kovetkeztetes --> [].
kovetkeztetes -->
	egyszerusitese(Valt),
	(   {Valt == valtozott} -> []
	;   valasztas
	),
	kovetkeztetes.

% A KL0 korlatlista KL-le egyszerusitheto (beleertve a valtozatlanul
% hagyast), valasztasi pont letrehozasa nelkul. Valt == valtozott csakkor,
% ha KL nem azonos KL0-val.
% :- pred egyszerusitese(atom, list(korlat)::in, list(korlat)::out).
egyszerusitese(Valt, KL0, KL) :-
	egyszerusitese(KL0, Valt, KL, []).

% egyszerusitese(KL, Valt, EKL0, EKL):
% A KL korlatlista az EKL0-EKL kulonbseglista altal leirt korlatlistava
% egyszerusitheto, valasztasi pont letrehozasa nelkul. Valt == valtozott
% csakkor, ha legalabb egy egyszerusites volt.
% :- pred egyszerusitese(list(korlat)::in, atom,
%                         list(korlat)::out, list(korlat)::in).
egyszerusitese([], _) --> [].
egyszerusitese([K|KL], Valt) -->
	egyszerusites(K, Valt),
	egyszerusitese(KL, Valt).

% egyszerusites(K, Valt, EKL0, EKL):
% A K korlat az EKL0-EKL kulonbseglista altal leirt korlatlistava
% egyszerusitheto, valasztasi pont letrehozasa nelkul. Valt == valtozott
% csakkor, ha volt egyszerusites.
% :- pred egyszerusites(korlat::in, atom,
%                       list(korlat)::out, list(korlat)::in).
egyszerusites(sum(L, Sum), Valt) -->
	sum_egyszerusites(L, Sum, Valt).
egyszerusites(csum(P, L, Sum), Valt) -->
	csum_egyszerusites(P, L, Sum, Valt).
egyszerusites(edge(P, Q, PK, QK), Valt) -->
	edge_egyszerusites(P, Q, PK, QK, Valt).

% sum_egyszerusites(L0, Sum0, Valt, EKL0, EKL):
% A sum(L0,Sum0) korlat az EKL0-EKL kulonbseglista altal leirt
% korlatlistava egyszerusitheto, valasztasi pont letrehozasa nelkul.
% Valt == valtozott csakkor, ha volt egyszerusites.
% :- pred sum_egyszerusites(list(kpont)::in, int::in, atom,
%                           list(korlat)::out, list(korlat)::in).
sum_egyszerusites(L0, Sum0, Valt) -->
	{osszeg_egyszerusites(L0, Sum0, L, Sum, Valt)},
	{length(L, Len)},
	(   {Len == Sum} -> {mind_azonos(L, 1)}
	;   {Sum == 0} -> {mind_azonos(L, 0)}
	;   {Sum > 0, Sum < Len},  [sum(L, Sum)]
	).

% osszeg_egyszerusites(L0, S0, L, S, Valt):
% Az L0 listabol elhagyva a behelyettesitett elemeket kapjuk L-t, S0-bol
% kivonva az elhagyott elemek erteket kapjuk S-t. Valt == valtozott
% csakkor, ha volt elhagyas.
% :- pred osszeg_egyszerusites(list(kpont)::in, int::in,
%                              list(kpont)::out, int::out, atom).
osszeg_egyszerusites([], S, [], S, _).
osszeg_egyszerusites([X|Xk], S0, L, S, Valt) :-
	(   nonvar(X) ->
	    Valt = valtozott,
	    S1 is S0-X,
	    osszeg_egyszerusites(Xk, S1, L, S, Valt)
	;   L = [X|L1],
	    osszeg_egyszerusites(Xk, S0, L1, S, Valt)
	).

% mind_azonos(L, E):
% L minden eleme egyesitheto E-vel.
% :- pred mind_azonos(list(kpont)::in, kpont::in).
mind_azonos([], _).
mind_azonos([E|L], E) :-
	mind_azonos(L, E).

% csum_egyszerusites(P, L, Sum, Valt, EKL0, EKL):
% A csum(P,L,Sum) korlat az EKL0-EKL kulonbseglista altal leirt
% korlatlistava egyszerusitheto, valasztasi pont letrehozasa nelkul.
% Valt == valtozott csakkor, ha volt egyszerusites.
% :- pred csum_egyszerusites(kpont::in, list(kpont)::in, int::in, atom,
%                            list(korlat)::out, list(korlat)::in).
csum_egyszerusites(P, L, Sum, Valt) -->
	{P == 1}, !,
	{Valt = valtozott},
	sum_egyszerusites(L, Sum, Valt).
csum_egyszerusites(P, _L, _Sum, Valt) -->
	{P == 0}, !,
	{Valt = valtozott}.
csum_egyszerusites(P, L0, Sum0, Valt) -->
	{osszeg_egyszerusites(L0, Sum0, L, Sum, Valt)},
	{length(L, Len)},
	(   {Sum < 0; Sum > Len} ->
	    {P = 0},
	    {Valt = valtozott}
	;   [csum(P, L, Sum)]
	).

% edge_egyszerusites(P, Q, PK, QK, Valt, EKL0, EKL):
% A edge(P,Q,PK,QK) korlat az EKL0-EKL kulonbseglista altal leirt
% korlatlistava egyszerusitheto, valasztasi pont letrehozasa nelkul.
% Valt == valtozott csakkor, ha volt egyszerusites.
% :- pred edge_egyszerusites(kpont::in, kpont::in, var, var, atom,
%                            list(korlat)::out, list(korlat)::in).
edge_egyszerusites(P, Q, PK, QK, Valt) -->
	{kapcsolat_korlat_kiertekelheto(P, Q, PK, QK, Cel)}, !,
	{Valt = valtozott},
	{Cel}.
edge_egyszerusites(P, Q, PK, QK, _Valt) -->
	[edge(P, Q, PK, QK)].

% kapcsolat_korlat_kiertekelheto(P, Q, PK, QK, Cel):
% Az edge(P, Q, PK, QK) korlat igazsagerteke eldontheto a
% Cel cel segitsegevel.
% :- pred kapcsolat_korlat_kiertekelheto(kpont::in, kpont::in,
%                                        var, var, univ::out)
kapcsolat_korlat_kiertekelheto(P, _, _, _, Cel) :-
	P == 0, !, Cel = true.
kapcsolat_korlat_kiertekelheto(_, Q, _, _, Cel) :-
	Q == 0, !, Cel = true.
kapcsolat_korlat_kiertekelheto(P, Q, PK, QK, Cel) :-
	P == 1, Q == 1, Cel = osszekapcsol(PK, QK).

% PK es QK kapcsolati valtozok nem azonosak. Egyesitjuk oket.
% :- pred osszekapcsol(var, var)
osszekapcsol(PK, QK) :-
	PK \== QK, PK = QK.

% valasztas(KL0, KL): KL0-ban egy megfeleloen megvalasztott)
% pontvaltozo behelyettesittetik. KL = KL0.
% (A ket parameterre csak a DCG miatt van szukseg.)
% :- pred valasztas(list(korlat)::in, list(korlat)::out).
valasztas(KL, KL) :-
	legjobb_valtozo(KL, X), !,
	choicepoint_made,
	nullegy(X),
	choice_made.
valasztas(KL, KL) :-
	memberchk(csum(P,_,_), KL),
	choicepoint_made,
	nullegy(P),
	choice_made.

% nullegy(X): X 0 vagy 1
% :- pred nullegy(kpont).
nullegy(0).
nullegy(1).

% legjobb_valtozo(KL, X): A KL korlatlistaban a legkisebb merteku sum
% korlat (lasd sum_korlat_merteke) egy valtozoja X. A KL korlatlistarol
% tudjuk, hogy tovabb nem egyszerusitheto, ezert a sum korlat
% listaargumentuma biztosan csak valtozokat tartalmaz.
% :- pred legjobb_valtozo(list(korlat)::in, kpont).
legjobb_valtozo(KL, X) :-
	legjobb_valtozo(KL, 10000000, V, X),
	X \== V.

% legjobb_valtozo(KL, M0, X0, X): Ha a KL korlatlistaban a legkisebb
% merteku sum korlat merteke kisebb M0-nal, akkor X ennek egy valtozoja
% lesz, egyebkent X = X0.
% :- pred legjobb_valtozo(list(korlat)::in, kpont).
legjobb_valtozo([], _, X, X).
legjobb_valtozo([K|KL], M0, X0, X) :-
	sum_korlat_merteke(K, M1, X1), !,
	(   M1 < M0 ->
	    legjobb_valtozo(KL, M1, X1, X)
	;   legjobb_valtozo(KL, M0, X0, X)
	).
legjobb_valtozo([_|KL], M0, X0, X) :-
	legjobb_valtozo(KL, M0, X0, X).

% sum_korlat_merteke(K, M, X): K egy sum korlat, amelynek merteke M, es egy
% tetszoleges (esetunkben az elso) valtozoja X.  Egy korlat merteke azt
% mutatja, hogy minimalisan hany valtozo megfelelo behelyettesitesevel
% valik egyertelmuve.
% Peldak:
%          sum([X,Y,Z,U,V], 3) merteke 2,
% 		mert 2 valtozo 0-ra helyettesitesevel egyertelmuve valik.
%          sum([X,Y,Z,U,V], 2) merteke 2,
% 		mert 2 valtozo 1-re helyettesitesevel egyertelmuve valik.
% :- pred sum_korlat_merteke(korlat::in, int::in, kpont).
sum_korlat_merteke(sum(L,N), M, X) :-
	length(L, Len), M is min(N, Len-N), L = [X|_].


% Altalanos segedeljarasok
% ------------------------

%-------------------- Matrix operations

% :- type matrix(Type) == list(list(Type)).
% A matrix is represented by a list of rows, which are lists of
% elements. All rows must be of equal length.

% matrix_elem(I, J, M, E): E is the element of matrix M in row I, column J.
% :- pred matrix_elem(int::in, int::in, matrix(T)::in, T::out).
matrix_elem(RowI, ClmI, Matrix, Elem) :-
        nth(RowI, Matrix, Row),
        nth(ClmI, Row, Elem).

% Matrix is a matrix of size RowNum x ClmNum, with all elements uninstantiated.
% :- pred matrix_of_vars(int::in, int::in, matrix(T)::out).
matrix_of_vars(0, _ClmNum, []) :- !.
matrix_of_vars(RowNum, ClmNum, [Row|Rows]) :-
	length(Row, ClmNum),
	RowNum1 is RowNum-1,
	matrix_of_vars(RowNum1, ClmNum, Rows).

%-------------------- Miscellaneous

% between(I, J, X): I =< X, X =< J. Capable of enumerating all such Xs.
% :- pred between(int::in, int::in, int::out).
between(I, J, I) :- I =< J.
between(I, J, X) :-
	I < J, I1 is I+1,
	between(I1, J, X).




% Probafuttatas
% -------------

% A kiadott keretprogrambeli stopper/2 csekely modositasa,
% a visszalepesek szamlalasara (vo. a backtrack_count modullal is).
% stopper2(+FileIn, +FileOut):
% Mint megold/2, de a futashoz szukseges idot is kiirja.
stopper2(FileIn, FileOut) :-
	statistics(runtime, [T0|_]),
	clear_counter(btracks),
	megold(FileIn, FileOut),
	statistics(runtime, [T1|_]),
	T is T1 - T0,
	get_counter(btracks, BT),
	format('Feladat: ~a, futsi id: ~3d s, ~w visszalps.~n',
	       [FileIn,T,BT]).

probafuttatas :-
	stopper2('tests3/test1d.txt',  'results3/test1r.txt'),
	stopper2('tests3/test2d.txt',  'results3/test2r.txt'),
	stopper2('tests3/test3d.txt',  'results3/test3r.txt'),
	stopper2('tests3/test4d.txt',  'results3/test4r.txt'),
	stopper2('tests3/test5d.txt',  'results3/test5r.txt'),
	stopper2('tests3/test6d.txt',  'results3/test6r.txt'),
	stopper2('tests3/test7d.txt',  'results3/test7r.txt'),
	stopper2('tests3/test8d.txt',  'results3/test8r.txt'),
	stopper2('tests3/test9d.txt',  'results3/test9r.txt'),
	stopper2('tests3/test10d.txt', 'results3/test10r.txt'),
	stopper2('tests3/test11d.txt', 'results3/test11r.txt'),
	stopper2('tests3/test12d.txt', 'results3/test12r.txt'),
	stopper2('tests3/test13d.txt', 'results3/test13r.txt'),
	stopper2('tests3/test14d.txt', 'results3/test14r.txt'),
	stopper2('tests3/test15d.txt', 'results3/test15r.txt').
